/*-------------------------------------------------------------------------
 *
 * freelist.c
 *	  routines for manipulating the buffer pool's replacement strategy.
 *
 * The name "freelist.c" is now a bit of a misnomer, since this module
 * controls not only the list of free buffers per se, but the entire
 * mechanism for looking up existing shared buffers and the strategy
 * for choosing replacement victims when needed.
 *
 * Note: all routines in this file assume that the BufMgrLock is held
 * by the caller, so no synchronization is needed.
 *
 *
 * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
 * Portions Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *
 *-------------------------------------------------------------------------
 */
#include "postgres.h"

#include <time.h>

#include "access/xact.h"
#include "storage/buf_internals.h"
#include "storage/bufmgr.h"


/*
 * The shared strategy control information. We have provided a useful variable
 * for you to use in your coding of CLOCK. The current clock position is indicated
 * by the clock_position variable. Use this to keep track of your clock hand position.
 */
typedef struct
{
	Buffer	clock_position;		/* Current clock position */

	long	num_lookup;		/* Some hit statistics */
	long	num_hit;
	time_t	stat_report;
} BufferStrategyControl;

/* GUC variable: time in seconds between statistics reports */
int DebugSharedBuffers = 0;

/* Pointers to shared state */
static BufferStrategyControl *StrategyControl = NULL;

/* Backend-local state about whether currently vacuuming */
static bool          strategy_hint_vacuum = false;
static TransactionId strategy_vacuum_xid;

/*
 * Printout for use when DebugSharedBuffers is enabled
 */
static void
StrategyStatsDump(void)
{
  time_t now = time(NULL);

  if (StrategyControl->stat_report + DebugSharedBuffers < now)
  {
    long all_hit;
    int  id, clean;
    ErrorContextCallback *errcxtold;

    clean = 0;
    for (id = 0; id < NBuffers; id++)
      if (!(BufferDescriptors[id].flags & BM_DIRTY)) clean++;

    if (StrategyControl->num_lookup == 0)
      all_hit = 0;
    else
    {
      all_hit = (StrategyControl->num_hit * 100 / StrategyControl->num_lookup);
    }

    errcxtold = error_context_stack;
    error_context_stack = NULL;
    elog(DEBUG1, "Q HIT RATE: %4ld%%", all_hit);
    elog(DEBUG1, "Q clean buffers: %5d", clean);
    error_context_stack = errcxtold;

    StrategyControl->num_lookup = 0;
    StrategyControl->num_hit    = 0;
    StrategyControl->stat_report = now;
  }
}

/*
 * StrategyBufferLookup
 *
 *	Lookup a page request in the cache directory. A buffer is only
 *	returned for a cache hit. 
 *
 *	recheck indicates we are rechecking after I/O wait; do not change
 *	internal status in this case.
 *
 *	*found_index is set to the index of the found buffer block, or -1 if none.
 *	This is not intended to be used by the caller, except to pass to
 *	StrategyReplaceBuffer().
 */
BufferDesc *
StrategyBufferLookup(BufferTag *tagPtr, bool recheck, int *found_index)
{
  /* Optional stats printout */
  if (DebugSharedBuffers > 0) {
    StrategyStatsDump();
  }

  return NULL;
}


/*
 * StrategyGetBuffer
 *
 *	Called by the bufmgr to get the next candidate buffer to use in
 *	BufferAlloc(). The only hard requirement BufferAlloc() has is that
 *	this buffer must not currently be pinned.
 */
BufferDesc *
StrategyGetBuffer(int *replace_index)
{

  /* No unpinned buffers at all!!!  */
  elog(ERROR, "no unpinned buffers available");
  return NULL; 	/* not reached */
}


/*
 * StrategyReplaceBuffer
 *
 * Called by the buffer manager to inform us that it flushed a buffer
 * and is now about to replace the content. Basically, this is a bookkeeping
 * routine. 
 *
 *  TA Comment: As you can see, you may ignore found_index for this project.
 *              Also notice, that replace_index is what YOUR code returned in 
 *              StrategyGetBuffer.
 */
void
StrategyReplaceBuffer(BufferDesc *buf, BufferTag *newTag, 
                      int found_index, int replace_index)
{
  Assert(found_index < 0);
}


/*
 * StrategyInvalidateBuffer
 *
 *	Called by the buffer manager to inform us that a buffer content
 *	is no longer valid. We simply throw away any eventual existing
 *	buffer hash entry.
 */
void
StrategyInvalidateBuffer(BufferDesc *buf)
{
}

/* TA Comment: Ignore this routine.
 * StrategyHintVacuum -- tell us whether VACUUM is active
 */
void
StrategyHintVacuum(bool vacuum_active)
{
 strategy_hint_vacuum = vacuum_active;
 strategy_vacuum_xid = GetTopTransactionId();
}

/*
 * StrategyDirtyBufferList
 *
 * Returns a list of dirty buffers, in priority order for writing.
 * Note that the caller may choose not to write them all.
 *
 * The caller must beware of the possibility that a buffer is no longer dirty,
 * or even contains a different page, by the time he reaches it.  If it no
 * longer contains the same page it need not be written, even if it is (again)
 * dirty.
 *
 * Buffer pointers are stored into buffers[], and corresponding tags into
 * buftags[], both of size max_buffers.  The function returns the number of
 * buffer IDs stored.
 */
int
StrategyDirtyBufferList(BufferDesc **buffers, BufferTag *buftags, int max_buffers)
{
  return 0;
}


/* TA Comment: This is done for you.
 * StrategyShmemSize
 *
 * estimate the size of shared memory used by the freelist-related structures.
 */
int
StrategyShmemSize(void)
{
  int size  = 0;

  /* size of lookup hash table */
  size += BufTableShmemSize(NBuffers);

  /* size of the shared replacement strategy control block */
  size += MAXALIGN(sizeof(BufferStrategyControl));

  return size;
}

/* TA Comment: This is done for you.
 * StrategyInitialize -- initialize the buffer cache replacement strategy.
 */
void
StrategyInitialize(bool init)
{
  bool	found;

  /* Initialize the shared lookup hashtable */
  InitBufTable(NBuffers);

  /* Get or create the shared strategy control block */
  StrategyControl = (BufferStrategyControl *)
    ShmemInitStruct("Buffer Strategy Status", sizeof(BufferStrategyControl), &found);
  if (!found)
  {
    /* Only done once, usually in postmaster */
    Assert(init);
    StrategyControl->clock_position = -1;
    StrategyControl->num_hit        = 0;
    StrategyControl->num_lookup     = 0;
    StrategyControl->stat_report    = 0;
  }
  else Assert(!init);
}
